from __future__ import annotations

from .editor import NodeEditorObject, EclipseNodeEditor, NodeEditorTool
from PyQt6.QtGui import QColor
from PyQt6.QtCore import Qt
import math


class SocketConnection:
    def __init__(self):
        self.from_node = None
        self.from_socket = None
        self.to_node = None
        self.to_socket = None


class Socket(NodeEditorObject):
    def __init__(self):
        super().__init__()
        self.connected_height = 25
        self.width = 0
        self.height = 25
        self.pin_radius = 5
        self.pin_hover_radius = 7
        self.pin_color = QColor(255, 255, 255)
        self.pin_type = "input"
        self.connections: list[SocketConnection] = []
        self.is_hovered = False
        self.is_pin_hovered = False

    def render(self, node: Node, x: int, y: int, editor: EclipseNodeEditor):
        pin_x = x
        if self.pin_type == "input":
            pin_x = x

        if self.pin_type == "output":
            pin_x = x + node.width

        editor.draw_circle(pin_x, y + self.connected_height / 2, self.pin_hover_radius if self.is_pin_hovered else self.pin_radius,
                           self.pin_color.red() / 255,
                           self.pin_color.green() / 255,
                           self.pin_color.blue() / 255)

    def check_show_interface(self):
        if self.pin_type == "output":
            return False

        if self.pin_type == "input":
            if self.connections:
                return False

        return True


class Node(NodeEditorObject):
    def __init__(self, x=0, y=0):

        super().__init__()
        # Position and size
        self.x = x
        self.y = y
        self.width = 100
        self.height = 100

        # Colors
        self.primary_color = QColor(75, 75, 75)    # Header color
        self.secondary_color = QColor(50, 50, 50)  # Background color
        self.title_color = QColor(255, 255, 255)   # Title text color

        # Title properties
        self.title = "Node"
        self.title_font_size = 12
        self.header_size = 30

        # State
        self.is_selected = False
        self.is_hovered = False
        self.is_over_header = False
        self.start_mouse_x = 0
        self.start_mouse_y = 0
        self.start_x = 0
        self.start_y = 0
        self.is_selected = False

        # Sockets
        self.sockets: list[Socket] = []
        self.sockets_spacing = 2
        self.render_socket_regions = False

    def add_input_socket(self, socket: Socket):
        socket.pin_type = "input"
        self.sockets.append(socket)
        return socket

    def add_output_socket(self, socket: Socket):
        socket.pin_type = "output"
        self.sockets.append(socket)
        return socket

    def raycast_reset(self, editor: EclipseNodeEditor):
        self.is_hovered = False
        self.is_over_header = False
        for socket in self.sockets:
            socket.is_hovered = False
            socket.is_pin_hovered = False

    def raycast(self, editor: EclipseNodeEditor):
        mouse_x, mouse_y = editor.get_mouse_pos()
        has_hit = False

        # Check if the mouse is over the node
        if mouse_x >= self.x and mouse_x <= self.x + self.width and mouse_y >= self.y and mouse_y <= self.y + self.height:
            has_hit = True
            self.is_hovered = True

            if mouse_y <= self.y + self.header_size:
                self.is_over_header = True

        # Check if the mouse is over the sockets
        socket_y = self.y + self.header_size
        for i, socket in enumerate(self.sockets):
            if mouse_x >= self.x and mouse_x <= self.x + self.width and mouse_y >= socket_y and mouse_y <= socket_y + socket.height:
                socket.is_hovered = True
                has_hit = True

            dist = math.sqrt((mouse_x - self.get_socket_x(socket))
                             ** 2 + (mouse_y - (socket_y + socket.connected_height / 2)) ** 2)
            if dist <= socket.pin_hover_radius:
                socket.is_pin_hovered = True
                has_hit = True
            show_interface = socket.check_show_interface()
            if show_interface:
                socket_y += socket.height + self.sockets_spacing
            else:
                socket_y += socket.connected_height + self.sockets_spacing

        return has_hit

    def input(self, editor: EclipseNodeEditor):
        mouse_x, mouse_y = editor.get_mouse_pos()
        if editor.mouse_pressed:
            if self.is_over_header:
                self.start_mouse_x = mouse_x
                self.start_mouse_y = mouse_y
                self.start_x = self.x
                self.start_y = self.y
                self.is_selected = True
            else:
                if not editor.get_key(Qt.Key.Key_Shift):
                    self.is_selected = False

        if editor.mouse_released:
            self.start_mouse_x = mouse_x
            self.start_mouse_y = mouse_y
            self.start_x = self.x
            self.start_y = self.y

    def update(self, editor: EclipseNodeEditor):
        """Update node state"""
        width_needed = self.width
        height_needed = self.header_size
        for socket in self.sockets:
            width_needed = max(width_needed, socket.width)
            show_interface = socket.check_show_interface()
            if show_interface:
                height_needed += socket.height + self.sockets_spacing
            else:
                height_needed += socket.connected_height + self.sockets_spacing

        self.width = width_needed
        self.height = height_needed

    def render(self, editor: EclipseNodeEditor):
        """Render the node"""
        def render_pass_node():
            # Draw outer shadow first (larger and more transparent)
            outer_shadow_offset = 6
            outer_shadow_expand = 4
            editor.draw_rectangle(
                self.x - outer_shadow_expand,
                self.y + outer_shadow_offset - outer_shadow_expand - 6,
                self.width + outer_shadow_expand * 2,
                self.height + outer_shadow_expand * 4,
                0, 0, 0, 0.15  # Very transparent black
            )

            # Draw inner shadow (smaller and more opaque)
            inner_shadow_offset = 4
            inner_shadow_expand = 2
            editor.draw_rectangle(
                self.x - inner_shadow_expand,
                self.y + inner_shadow_offset - inner_shadow_expand - 4,
                self.width + inner_shadow_expand * 2,
                self.height + inner_shadow_expand * 4,
                0, 0, 0, 0.4  # Semi-transparent black

            )

            # Draw node background (main body)
            editor.draw_rectangle(
                self.x,
                self.y,
                self.width,
                self.height,
                self.secondary_color.red() / 255,
                self.secondary_color.green() / 255,
                self.secondary_color.blue() / 255
            )

            # Draw node header
            editor.draw_rectangle(
                self.x,
                self.y,
                self.width,
                self.header_size,
                self.primary_color.red() / 255,
                self.primary_color.green() / 255,
                self.primary_color.blue() / 255
            )

            if self.is_selected:
                editor.draw_rectangle_outline(
                    self.x - 15,
                    self.y - 15,
                    self.width + 30,
                    self.height + 30,
                    1, 1, 1, 0.2, 2
                )

            if self.is_over_header:
                editor.draw_rectangle(
                    self.x,
                    self.y,
                    self.width,
                    self.header_size,
                    1, 1, 1, 0.05
                )

            # Draw title text
            # Center the text vertically in the header
            text_x = self.x + 10  # Add some padding from the left
            text_y = self.y + (self.header_size / 2)  # Center point of header

            editor.draw_text(
                text_x,
                text_y,
                self.title,
                self.title_color.red() / 255,
                self.title_color.green() / 255,
                self.title_color.blue() / 255,
                1.0,
                self.title_font_size
            )

            # Render sockets
            socket_y = self.y + self.header_size
            for i, socket in enumerate(self.sockets):
                if socket.is_hovered:
                    editor.draw_rectangle(
                        self.x,
                        socket_y,
                        self.width,
                        socket.height,
                        1, 1, 1, 0.03
                    )
                if self.render_socket_regions:
                    clr_r = 1
                    clr_g = 0
                    clr_b = 0
                    if i % 2 == 1:
                        clr_r = 0
                        clr_g = 1
                        clr_b = 1
                    editor.draw_rectangle(
                        self.x,
                        socket_y,
                        self.width,
                        socket.height,
                        clr_r, clr_g, clr_b, 0.03
                    )

                socket.render(self, self.x, socket_y, editor)
                show_interface = socket.check_show_interface()
                if show_interface:
                    socket_y += socket.height + self.sockets_spacing
                else:
                    socket_y += socket.connected_height + self.sockets_spacing

        def render_pass_connections():
            for socket in self.sockets:
                for connection in socket.connections:
                    connection: SocketConnection
                    from_node = connection.from_node
                    to_node = connection.to_node

                    from_socket = connection.from_socket
                    to_socket = connection.to_socket

                    from_x, from_y = from_node.get_socket_pos(from_socket)
                    to_x, to_y = to_node.get_socket_pos(to_socket)

                    editor.draw_line(from_x, from_y, to_x, to_y, 2, 1, 1, 1, 1)

        editor.add_render_group(2, render_pass_node)
        editor.add_render_group(1, render_pass_connections)

    def get_socket_x(self, socket: Socket):
        if socket.pin_type == "input":
            return self.x
        if socket.pin_type == "output":
            return self.x + self.width

    def get_socket_pos(self, socket: Socket):
        socket_y = self.y + self.header_size
        for current_socket in self.sockets:
            if current_socket == socket:
                return self.get_socket_x(current_socket), socket_y + current_socket.connected_height / 2
            show_interface = current_socket.check_show_interface()
            if show_interface:
                socket_y += current_socket.height + self.sockets_spacing
            else:
                socket_y += current_socket.connected_height + self.sockets_spacing


class NodeMovingTool(NodeEditorTool):
    def __init__(self):
        super().__init__()

    def input(self, editor: EclipseNodeEditor):
        mouse_x, mouse_y = editor.get_mouse_pos()

        if editor.mouse_down:
            nodes = editor.get_objects_of_type(Node)
            if editor.grabbed_object:
                if isinstance(editor.grabbed_object, Node):
                    for obj in nodes:
                        obj: Node
                        if obj.is_selected:
                            obj.x = obj.start_x + mouse_x - editor.grabbed_object.start_mouse_x
                            obj.y = obj.start_y + mouse_y - editor.grabbed_object.start_mouse_y


class NodeConnectionTool(NodeEditorTool):
    def __init__(self):
        super().__init__()
        self.from_node_and_socket: tuple[Node, Socket] | None = None
        self.to_node_and_socket: tuple[Node, Socket] | None = None

    def input(self, editor: EclipseNodeEditor):
        mouse_x, mouse_y = editor.get_mouse_pos()

        if editor.mouse_pressed:
            nodes = editor.get_objects_of_type(Node)
            self.from_node_and_socket = None
            for node in reversed(nodes):
                node: Node
                for socket in node.sockets:
                    if socket.is_pin_hovered:
                        self.from_node_and_socket = (node, socket)
                        break

        if editor.mouse_released:
            nodes = editor.get_objects_of_type(Node)
            if self.from_node_and_socket:
                from_node, from_socket = self.from_node_and_socket
                for node in reversed(nodes):
                    node: Node
                    for socket in node.sockets:
                        if socket.is_pin_hovered:
                            self.to_node_and_socket = (node, socket)
                            break

                if self.to_node_and_socket:
                    to_node, to_socket = self.to_node_and_socket
                    if from_node != to_node:
                        if from_socket.pin_type == "input" and to_socket.pin_type == "output":
                            self.create_input_output_connection(
                                from_node, from_socket, to_node, to_socket)
                        elif from_socket.pin_type == "output" and to_socket.pin_type == "input":
                            self.create_output_input_connection(
                                to_node, to_socket, from_node, from_socket)

                self.from_node_and_socket = None
                self.to_node_and_socket = None

    def render(self, editor: EclipseNodeEditor):
        def render_temp_connection():
            if self.from_node_and_socket:
                mouse_x, mouse_y = editor.get_mouse_pos()
                from_node, from_socket = self.from_node_and_socket
                socket_x, socket_y = from_node.get_socket_pos(from_socket)
                editor.draw_line(socket_x, socket_y, mouse_x,
                                 mouse_y, 2, 1, 1, 1, 1)

        editor.add_render_group(10, render_temp_connection)

    def create_input_output_connection(self, from_node: Node, from_socket: Socket, to_node: Node, to_socket: Socket):
        connection = SocketConnection()
        connection.from_node = from_node
        connection.from_socket = from_socket
        connection.to_node = to_node
        connection.to_socket = to_socket
        from_socket.connections.append(connection)
        to_socket.connections.append(connection)

    def create_output_input_connection(self, from_node: Node, from_socket: Socket, to_node: Node, to_socket: Socket):
        connection = SocketConnection()
        connection.from_node = from_node
        connection.from_socket = from_socket
        connection.to_node = to_node
        connection.to_socket = to_socket
        from_socket.connections.append(connection)
        to_socket.connections.append(connection)
